
#include <erl_driver.h>
#include <ei.h>

/* Driver interface declarations */

static int init(void);
static ErlDrvData start(ErlDrvPort port, char *command);
static void stop(ErlDrvData drv_data);
static void finish(void);
static void output(ErlDrvData drv_data, char *buf, int len);
static void ready_async(ErlDrvData drv_data, ErlDrvThreadData async_data);

static ErlDrvEntry test_driver_entry = {
    init,			/* init */
    start,			/* start */
    stop,			/* stop */
    output,			/* output */
    NULL,			/* ready_input */
    NULL,			/* ready_output */ 
    "drv",			/* the name of the driver */
    finish,			/* finish */
    NULL,			/* handle */
    NULL,			/* control */
    NULL,			/* timeout */
    NULL,			/* outputv */
    ready_async,		/* ready_async */
    NULL,			/* flush */
    NULL,			/* call */
    NULL			/* event */
};

/* INITIALIZATION AFTER LOADING */

/* 
 * This is the init function called after this driver has been loaded.
 * It must *not* be declared static. Must return the address to 
 * the driver entry.
 */

DRIVER_INIT(test_drv)
{
    return &test_driver_entry;
}

// This is called directly after the driver has been loaded by erl_ddll:load_driver/2
//
static int init(void)
{
    fprintf(stderr, "init called from %0X\r\n", pthread_self());
}

typedef struct {
    ErlDrvPort port;
} my_drv_data;

static ErlDrvData start(ErlDrvPort port, char *command)
{ 
    fprintf(stderr, "start(%0X, %s) called from\r\n", port, command, pthread_self());
    if (port == NULL) {
        return ERL_DRV_ERROR_GENERAL;
    }
    my_drv_data *data = driver_alloc (sizeof(my_drv_data));
    if (data == NULL) {
        return ERL_DRV_ERROR_GENERAL;
    }
    data->port = port;
    return (ErlDrvData) data;
}

static void stop(ErlDrvData drv_data)
{
    fprintf(stderr, "stop(%0X) called from thread %0X\r\n", drv_data, pthread_self());
    if (drv_data == 0) return;
    fprintf(stderr, "port used by stop(): %0X\r\n", ((my_drv_data *) drv_data)->port);
    // driver_free (drv_data);
}

static void finish()
{
    fprintf(stderr, "finish() called from thread %0X\r\n", pthread_self());
}

static void do_invoke(void *data)
{
    fprintf(stderr, "do_invoke(%0X) called from thread %0X\r\n", data, pthread_self());
}

static void do_free(void *data)
{
    fprintf(stderr, "do_free(%0X) called from thread %0X\r\n", data, pthread_self());
}

static void ready_async(ErlDrvData drv_data, ErlDrvThreadData async_data)
{
    fprintf(stderr, "ready_async(%0X, %0X) called from thread %0X\r\n", drv_data, async_data, pthread_self());
    if (drv_data == 0) return;
    ErlDrvPort port = ((my_drv_data *) drv_data)->port;
    fprintf(stderr, "port used by ready_async(): %0X\r\n", port);
    ei_x_buff *test = driver_alloc (sizeof (ei_x_buff));
    ei_x_new_with_version (test);
    ei_x_encode_atom (test, "aaa");
    driver_output(port, test->buff, test->index);
}

static void output(ErlDrvData drv_data, char *buf, int len)
{
    fprintf(stderr, "output(%0X, %0X, %d) called from thread %0X\r\n", drv_data, buf, len, pthread_self());
    if (len <= 0) return;
    if (drv_data == 0) return;
    ErlDrvPort port = ((my_drv_data *) drv_data)->port;
    fprintf(stderr, "port used by output(): %0X\r\n", port);
    driver_async (port, NULL, do_invoke, NULL, do_free);
}
